/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`define ALU_STAND_ALONE  // for CPLD ALU (XC2C64A-VQ44)
`ifdef ALU_STAND_ALONE
  `define ALU_EXTENDED
  `define ALU_WITH_ACC
//  `define ALU_ES1_ATOB // too large for XC2C64A
  `define ALU_SIO_TRI    // shrink pins for VQ44  (for top-level only)
`elsif ALU_FOR_ZSP8	 // for zsp8 processor
  `define ALU_EXTENDED
  `define ALU_WITH_ACC
  `define ALU_ES1_ATOB
//  `define ALU_SIO_TRI
`else
`include "rtavr_defs.v"
`endif
//
// Module Name: rtavr_alu
//   ver 002 (generic)
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
//                                        all-in-one  
// Number of Slice Flip Flops:              xxx
// Number of 4 input LUTs:                  xxx
// Number of occupied Slices:               xxx
// Total Number of 4 input LUTs:            xxx
// Number of bonded IOBs :                  xxx
//   IOB Flip Flops :                       xxx
// Number of BUFGMUXs :                     xxx
// Number used for Dual Port RAMs           xxx
// Number of RAMB16BWEs :                   xxx
// Clock to Setup 

module rtavr_alu # (
	parameter ALU_WIDTH = 8
	) (
		input [ALU_WIDTH-1:0] B
		, output [ALU_WIDTH-1:0] F
  `ifdef ALU_WITH_ACC
  `else
		, input [ALU_WIDTH-1:0] A
  `endif

		, input C_in
		, input Z_in
		, output C_out
		, output H_out
		, output V_out
		, output Z_out

		, input [2:0] S
`ifdef ALU_EXTENDED
		, input [1:0] ES
`endif
`ifdef ALU_WITH_ACC
		, input   [1:0] RS   // Reg. Select
		, input   CP         // Clock 
    `ifdef ALU_SIO_TRI
		, inout   SIO0       // RI/LO
		, inout   SIO3       // LI/RO
    `else
		, input   SI0        // RI/LO
		, output  SO0        // RI/LO
		, input   SI3        // LI/RO
		, output  SO3        // LI/RO
    `endif
`endif
		);

`ifdef ALU_WITH_ACC
    wire [ALU_WIDTH-1:0] A;
    reg [ALU_WIDTH-1:0] acc = 0;
    assign A[ALU_WIDTH-1:0] = acc[ALU_WIDTH-1:0];
`endif
// S == 0 : ADD (wo carry)
// S == 1 : ADC (with carry)
// S == 2 : SUB (wo carry)
// S == 3 : SBC (with carry)
// S == 4 : AND
// S == 5 : XOR
// S == 6 : OR
// S == 7 : MOV
//
//                      S = 0  1  2  3  4  5  6  7
// C_out  : Carry[8]        o  o  o  o  x  x  x  x
// H_out  : Carry[4]        o  o  o  o  x  x  x  x
// V_out  : Over Flow       o  o  o  o  0  0  0  x
// Z_out  : Zero            o  o  o (o) -  -  -  x
//          (o) if (Z_in == 1)
//           - no change
//           x  don't care
// EXTENDED Ops
//   ES[0] : load -1 to B
//   ES[1] : load  0 to A
//
// OP3          --*-:**------------                               HSVNZC
// CP/CPC       000B:01rd:dddd:rrrr	# B=1: CP  B=0: CPC   #   XXXXXX
// SUB/SBC	000B:10rd:dddd:rrrr	# B=1: sub B=0: sbc   #   XXXXXX
// ADD/ADD	000C:11rd:dddd:rrrr	# C=0: add C=1: adc   #   XXXXXX
// AND		0010:00rd:dddd:rrrr	# logical AND         #   -X0XX-
// EOR		0010:01rd:dddd:rrrr	# exclusive OR        #   -X0XX-
// OR		0010:10rd:dddd:rrrr	# inclusive OR        #   -X0XX-
// MOV		0010:11rd:dddd:rrrr	# copy register       #   ------
//
// COM             1001:010d:dddd:0000  #                     #   -X0XX1
// NEG             1001:010d:dddd:0001  #                     #   XXXXXX
// INC             1001:010d:dddd:0011  #                     #   -XXXX-
// DEC             1001:010d:dddd:1010  #                     #   -XXXX-
// NEG : 0 - B (A == B) : S = 2, ES[1] = 1
// COM : A ^ -1         : S = 5, ES[0] = 1
// INC : A - -1         : S = 2, ES[0] = 1
// DEC : A + -1         : S = 0, ES[0] = 1

    wire WC = (S[0] & C_in);
//    (~S[1] & ~WC)  : ADD or ADC(C==0)
//    (~S[1] &  WC)  : ADC(C==1)
//    ( S[1] & ~WC)  : SUB or SBC(C==0)
//    ( S[1] &  WC)  : SBC(C==1)

`ifdef ALU_EXTENDED
     wire [ALU_WIDTH-1:0] AX = (ES[1]) ? 0 : A;
  `ifdef ALU_ES1_ATOB
     wire [ALU_WIDTH-1:0] BX = (ES[1]) ? A : (ES[0]) ? -1 : B;
  `else
     wire [ALU_WIDTH-1:0] BX = (ES[0]) ? -1 : B;
  `endif
`else
     wire [ALU_WIDTH-1:0] AX = A;
     wire [ALU_WIDTH-1:0] BX = B;
`endif
     wire [ALU_WIDTH:0] WCX = WC; // change bit width
     wire [ALU_WIDTH:0] ALU_OUT = ((S == 0)|(S == 1))
			   ? ({1'b0, AX} + ({1'b0, BX } + WCX))
		     : ((S == 2)|(S == 3))
			   ? ({1'b0, AX} - ({1'b0, BX } + WCX))
		     : (S == 4) ? {1'b0, (AX & BX)}
		     : (S == 5) ? {1'b0, (AX ^ BX)}
		     : (S == 6) ? {1'b0, (AX | BX)}
		     : {1'b0, BX } ;

     assign Z_out = (S == 3) ? (Z_in & (ALU_OUT[ALU_WIDTH-1:0] == 0))
		  : (ALU_OUT[ALU_WIDTH-1:0] == 0);

     assign C_out = ALU_OUT[ALU_WIDTH];
     
     assign F = ALU_OUT[ALU_WIDTH-1:0];

      //  ((S == 4)|(S == 5)|(S == 6)) ? 0 // (S ==7) Not care
     assign V_out = (S[2]) ? 1'b0 :
	     ( ~(S[1]) & (F[ALU_WIDTH-1] & ~AX[ALU_WIDTH-1] & ~BX[ALU_WIDTH-1]
			 | ~F[ALU_WIDTH-1] &  AX[ALU_WIDTH-1] &  BX[ALU_WIDTH-1])
	      | (S[1]) & (F[ALU_WIDTH-1] & ~AX[ALU_WIDTH-1] &  BX[ALU_WIDTH-1]
			 | ~F[ALU_WIDTH-1] &  AX[ALU_WIDTH-1] & ~BX[ALU_WIDTH-1]) )
		;

     assign H_out = 
	      (	~(S[1]) & ( ~F[(ALU_WIDTH-1>>1)] & A[(ALU_WIDTH-1>>1)]
			 | ~F[(ALU_WIDTH-1>>1)] & BX[(ALU_WIDTH-1>>1)]
			 |  AX[(ALU_WIDTH-1>>1)] & BX[(ALU_WIDTH-1>>1)] )
	       | (S[1]) & ( F[(ALU_WIDTH-1>>1)] & ~AX[(ALU_WIDTH-1>>1)]
			 |  F[(ALU_WIDTH-1>>1)] & BX[(ALU_WIDTH-1>>1)]
			 | ~AX[(ALU_WIDTH-1>>1)] & BX[(ALU_WIDTH-1>>1)] ) )
		;
`ifdef ALU_WITH_ACC
 `ifdef ALU_SIO_TRI
    wire SO0;
    wire SO3;
    wire SI0;
    wire SI3;
 `endif
    assign SO0 =  F[0];
    assign SO3 = F[ALU_WIDTH-1];

    always @(posedge CP)
    begin
	acc  <=   (RS == 0) ? F[ALU_WIDTH-1:0]
		: (RS == 1) ? { F[ALU_WIDTH-2:0] , SI0 }
		: (RS == 2) ? { SI3 , F[ALU_WIDTH-2:0] }
		: acc;
    end

  `ifdef ALU_SIO_TRI
    assign SIO0 = (RS == 2) ? SO0 : 1'bZ ;
    assign SIO3 = (RS == 1) ? SO3 : 1'bZ ;

    assign SI0 = SIO0;
    assign SI3 = SIO3;
  `endif
`endif // ALU_WITH_ACC

endmodule
