module zsp8_core (
		input RESET
		, input CLK
		, input [7:0] BUS_DI
		, output [7:0] BUS_DO
		, output [15:0] BUS_ADDR
		, output BUS_RD
		, output BUS_WR
	);

    reg [5:0]   UPC = 6'h00;
    reg [2:0]   INST_ATTR = 3'h0;
    reg [23:0]  UROM [0:63]; 

    reg [7:0]   LATCH = 8'h00;
    reg [15:0]  PC = 16'h0000;
    reg [15:0]  INDEX = 16'h0000;

    wire [7:0]  load_data = BUS_DI[7:0];
    wire [7:0]  store_data;
    
    wire [23:0] urom_out = UROM[UPC][23:0];

    wire u_loadpc = urom_out[7];
    wire u_jump   = urom_out[6];
    wire u_br;
    always @(negedge CLK)
    begin
       if (RESET) 
          begin
             UPC <= 0;
          end 
       else
          begin
             if (u_loadpc)
	        begin
		   UPC <= { load_data[7:3] , 1'b0 };
		   INST_ATTR <= load_data[2:0];
		end
             else if (u_jump | u_br) UPC <= urom_out[5:0];
             else                    UPC <= UPC + 6'h1;
          end
    end


    wire c_bus_rd       = urom_out[8];
    wire c_bus_wr       = urom_out[9];
    wire c_addr_sel     = urom_out[10];  // 0: PC , 1: INDEX
    wire c_load_latch   = urom_out[11];
    wire c_load_index_l = urom_out[12];
    wire c_load_index_h = urom_out[13];
    wire c_load_pc_l    = urom_out[14];
    wire c_load_pc_h    = urom_out[15];
    wire c_inc_pc       = urom_out[16];

    assign BUS_ADDR = (c_addr_sel) ? INDEX[15:0]: PC[15:0];
    assign BUS_DO   = store_data;
    assign BUS_WR   = c_bus_wr;
    assign BUS_RD   = c_bus_rd;
    
    always @(negedge CLK) 
    begin
       if (RESET) 
          begin
             LATCH <= 0;
             PC    <= 0;
             INDEX <= 0;
          end 
       else
          begin
             if (c_load_latch)     LATCH[7:0]  <= load_data[7:0];
             if (c_load_index_l)   INDEX[7:0]  <= load_data[7:0];
             if (c_load_index_h)   INDEX[15:8] <= load_data[7:0];
             if (c_load_pc_l)      PC[7:0]  <= load_data[7:0];
             else if (c_load_pc_h) PC[15:8] <= load_data[7:0];
             else if (c_inc_pc)    PC <= PC + 16'h1;
          end
    end 

    reg C = 0;
    reg Z = 0;
    reg V = 0;

    wire [7:0] alu_f; 
    wire [7:0] alu_b = LATCH;

    assign store_data = alu_f[7:0];
    assign u_br       =   urom_out[21] 
    			      & ( ~V & (INST_ATTR[2:0] == 6)
			        | ~Z & (INST_ATTR[2:0] == 5) 
				| ~C & (INST_ATTR[2:0] == 4)
			        |  V & (INST_ATTR[2:0] == 2)
			        |  Z & (INST_ATTR[2:0] == 1) 
				|  C & (INST_ATTR[2:0] == 0) ); 
                                   
    wire [1:0] alu_rs   = urom_out[20:19];
    wire [1:0] alu_es   = urom_out[18:17];
    wire [2:0] alu_s    = INST_ATTR[2:0];

    rtavr_alu ALU (
		.B(alu_b), .F(alu_f)
		, .S(alu_s) , .ES(alu_es)
		, .CP(CLK) , .RS(alu_rs)
		, .C_in(C) , .Z_in(Z)
		, .C_out(C_OUT), .Z_out(Z_OUT)
		, .V_out(V_OUT) 
		, .SI0(SI0) , .SI3(SI3)
		, .SO0(SO0) , .SO3(SO3)
		);

    assign SHIFT_DATA =  (urom_out[23:22] == 3) ? SO3
			:(urom_out[23:22] == 2) ? SO0
			:(urom_out[23:22] == 1) ? C
			: 1'b0 ;
    assign SI0 = SHIFT_DATA;
    assign SI3 = SHIFT_DATA;
    
    always @(negedge CLK) 
    begin
       if (RESET) 
          begin
             C   <= 0;
             Z   <= 0;
	     V   <= 0;
          end 
       else
          begin
		// : (RS == 1) ? { F[ALU_WIDTH-2:0] , SI0 }
		// : (RS == 2) ? { SI3 , F[ALU_WIDTH-2:0] }
             C   <= (alu_rs == 1) ? SO3 
	           :(alu_rs == 2) ? SO0
	           : C_OUT;
             Z   <= Z_OUT;
	     V   <= V_OUT;
          end
    end 

    initial 
      begin
`ifdef ZZZ
	UROM[  0] <= 24'b000000000000000000000000;
	UROM[  1] <= 24'b000000000000000000000000;
	UROM[  2] <= 24'b000000000000000000000000;
	UROM[  3] <= 24'b000000000000000000000000;
	UROM[  4] <= 24'b000000000000000000000000;
	UROM[  5] <= 24'b000000000000000000000000;
	UROM[  6] <= 24'b000000000000000000000000;
	UROM[  7] <= 24'b000000000000000000000000;
	UROM[  8] <= 24'b000000000000000000000000;
	UROM[  9] <= 24'b000000000000000000000000;
	UROM[ 10] <= 24'b000000000000000000000000;
	UROM[ 11] <= 24'b000000000000000000000000;
	UROM[ 12] <= 24'b000000000000000000000000;
	UROM[ 13] <= 24'b000000000000000000000000;
	UROM[ 14] <= 24'b000000000000000000000000;
	UROM[ 15] <= 24'b000000000000000000000000;
	UROM[ 16] <= 24'b000000000000000000000000;
	UROM[ 17] <= 24'b000000000000000000000000;
	UROM[ 18] <= 24'b000000000000000000000000;
	UROM[ 19] <= 24'b000000000000000000000000;
	UROM[ 20] <= 24'b000000000000000000000000;
	UROM[ 21] <= 24'b000000000000000000000000;
	UROM[ 22] <= 24'b000000000000000000000000;
	UROM[ 23] <= 24'b000000000000000000000000;
	UROM[ 24] <= 24'b000000000000000000000000;
	UROM[ 25] <= 24'b000000000000000000000000;
	UROM[ 26] <= 24'b000000000000000000000000;
	UROM[ 27] <= 24'b000000000000000000000000;
	UROM[ 28] <= 24'b000000000000000000000000;
	UROM[ 29] <= 24'b000000000000000000000000;
	UROM[ 30] <= 24'b000000000000000000000000;
	UROM[ 31] <= 24'b000000000000000000000000;
	UROM[ 32] <= 24'b000000000000000000000000;
	UROM[ 33] <= 24'b000000000000000000000000;
	UROM[ 34] <= 24'b000000000000000000000000;
	UROM[ 35] <= 24'b000000000000000000000000;
	UROM[ 36] <= 24'b000000000000000000000000;
	UROM[ 37] <= 24'b000000000000000000000000;
	UROM[ 38] <= 24'b000000000000000000000000;
	UROM[ 39] <= 24'b000000000000000000000000;
	UROM[ 40] <= 24'b000000000000000000000000;
	UROM[ 41] <= 24'b000000000000000000000000;
	UROM[ 42] <= 24'b000000000000000000000000;
	UROM[ 43] <= 24'b000000000000000000000000;
	UROM[ 44] <= 24'b000000000000000000000000;
	UROM[ 45] <= 24'b000000000000000000000000;
	UROM[ 46] <= 24'b000000000000000000000000;
	UROM[ 47] <= 24'b000000000000000000000000;
	UROM[ 48] <= 24'b000000000000000000000000;
	UROM[ 49] <= 24'b000000000000000000000000;
	UROM[ 50] <= 24'b000000000000000000000000;
	UROM[ 51] <= 24'b000000000000000000000000;
	UROM[ 52] <= 24'b000000000000000000000000;
	UROM[ 53] <= 24'b000000000000000000000000;
	UROM[ 54] <= 24'b000000000000000000000000;
	UROM[ 55] <= 24'b000000000000000000000000;
	UROM[ 56] <= 24'b000000000000000000000000;
	UROM[ 57] <= 24'b000000000000000000000000;
	UROM[ 58] <= 24'b000000000000000000000000;
	UROM[ 59] <= 24'b000000000000000000000000;
	UROM[ 60] <= 24'b000000000000000000000000;
	UROM[ 61] <= 24'b000000000000000000000000;
	UROM[ 62] <= 24'b000000000000000000000000;
	UROM[ 63] <= 24'b000000000000000000000000;
`else  // random pattern for implement test
		UROM[  0] <= 24'hC010C0;
		UROM[  1] <= 24'hC029C0;
		UROM[  2] <= 24'hC027C0;
		UROM[  3] <= 24'hC025C0;
		UROM[  4] <= 24'hC023C0;
		UROM[  5] <= 24'hC021C0;
		UROM[  6] <= 24'hC01FC0;
		UROM[  7] <= 24'hC01DC0;
		UROM[  8] <= 24'hC01B27;
		UROM[  9] <= 24'hBF1FE3;
		UROM[ 10] <= 24'hE0D1BF;
		UROM[ 11] <= 24'hBFCDE0;
		UROM[ 12] <= 24'hE6A0E0;
		UROM[ 13] <= 24'hE0E2E4;
		UROM[ 14] <= 24'hC00291;
		UROM[ 15] <= 24'h932D36;
		UROM[ 16] <= 24'h07B1F7;
		UROM[ 17] <= 24'hE010E6;
		UROM[ 18] <= 24'hE0B0C0;
		UROM[ 19] <= 24'h933D37;
		UROM[ 20] <= 24'h07B1F7;
		UROM[ 21] <= 24'hD002C2;
		UROM[ 22] <= 24'hCFD393;
		UROM[ 23] <= 24'h931F93;
		UROM[ 24] <= 24'h93DFE6;
		UROM[ 25] <= 24'hE0D0EF;
		UROM[ 26] <= 24'h2F6C2F;
		UROM[ 27] <= 24'hD0DDE0;
		UROM[ 28] <= 24'h2F6C2F;
		UROM[ 29] <= 24'hD0D9EF;
		UROM[ 30] <= 24'h2F6C2F;
		UROM[ 31] <= 24'hD0DCE0;
		UROM[ 32] <= 24'h2F6C2F;
		UROM[ 33] <= 24'hD0D8E0;
		UROM[ 34] <= 24'hE0612F;
		UROM[ 35] <= 24'h2F5DD0;
		UROM[ 36] <= 24'hE081E0;
		UROM[ 37] <= 24'h2F4C2F;
		UROM[ 38] <= 24'hD0D5E0;
		UROM[ 39] <= 24'hE0612F;
		UROM[ 40] <= 24'h2F5DD0;
		UROM[ 41] <= 24'hE081E0;
		UROM[ 42] <= 24'h2F4C2F;
		UROM[ 43] <= 24'hD0D8E0;
		UROM[ 44] <= 24'hE060D0;
		UROM[ 45] <= 24'hE080E0;
		UROM[ 46] <= 24'hD0E2E0;
		UROM[ 47] <= 24'hE061D0;
		UROM[ 48] <= 24'hE080E0;
		UROM[ 49] <= 24'hD0DCE0;
		UROM[ 50] <= 24'hE060D0;
		UROM[ 51] <= 24'hE080E0;
		UROM[ 52] <= 24'hD0D9E0;
		UROM[ 53] <= 24'hE061D0;
		UROM[ 54] <= 24'hE080E0;
		UROM[ 55] <= 24'hD0D3E0;
		UROM[ 56] <= 24'hE060D0;
		UROM[ 57] <= 24'hE080E0;
		UROM[ 58] <= 24'hD0D0E0;
		UROM[ 59] <= 24'hE061D0;
		UROM[ 60] <= 24'hE080E0;
		UROM[ 61] <= 24'hD0CAEF;
		UROM[ 62] <= 24'hD0BFE0;
		UROM[ 63] <= 24'hD0BDEF;
`endif
      end
endmodule
